/* Copyright (C) 2002-2009 The SSI Project. All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/

#ifdef HAVE_CONFIG_H
        #include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
        #include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

char *lis_storagename2[]   = {"CRS", "CCS", "MSR", "DIA", "ELL", "JDS", "BSR", "BSC", "VBR", "COO", "DNS"};

#undef __FUNC__
#define __FUNC__ "main"
int main(int argc, char* argv[])
{
  LIS_MATRIX		A,A0;
  LIS_VECTOR		b,x,v;
  LIS_SCALAR		ntimes,nmflops,nnrm2;
  LIS_SCALAR		*value;

  int			nprocs,my_rank;
  int			gn,nnz,mode;
  int			i,j,jj,j0,j1,l,k,n,np,h,ih;
  int			m,nn,ii;
  int			bnr,bnc,cjdbnr;
  int			rn,rmin,rmax,rb;
  int			is,ie,clsize,ci,*iw;
  int			err,iter,storage;
  int	       		*ptr,*index;
  double		mem,val,ra,rs,ri,ria,ca,time,time2,convtime,val2,nnzs,nnzap,nnzt;
  double		commtime[2],comptime[2],flops[2];
  FILE			*file;
  char path[1024];

  LIS_DEBUG_FUNC_IN;
    
  lis_initialize(&argc, &argv);

#ifdef USE_MPI
  MPI_Comm_size(MPI_COMM_WORLD,&nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
  nprocs  = 1;
  my_rank = 0;
#endif

  if( argc < 4 )
    {
      if( my_rank==0 ) printf("Usage: spmvtest3 matrix_filename matrix_type iter\n");
      lis_finalize();
      exit(0);
    }

  storage  = atoi(argv[2]);

  iter = atoi(argv[3]);
  if( iter<=0 )
    {
      printf("iter=%d <= 0\n",iter);
      CHKERR(1);
    }

  if( my_rank==0 )
    {
      printf("\n");
      printf("number of processes = %d\n",nprocs);
    }

#ifdef _OPENMP
  if( my_rank==0 )
    {
      printf("max number of threads = %d\n",omp_get_num_procs());
      printf("number of threads = %d\n",omp_get_max_threads());
    }
#endif

  /* create matrix and vectors */
  lis_matrix_create(LIS_COMM_WORLD,&A0);
  err = lis_input(A0,NULL,NULL,argv[1]);
  CHKERR(err);

  n   = A0->n;
  gn  = A0->gn;
  nnz = A0->nnz;
  np  = A0->np-n;
#ifdef USE_MPI
  MPI_Allreduce(&nnz,&i,1,MPI_INT,MPI_SUM,A0->comm);
  nnzap = (double)i / (double)nprocs;
  nnzt  = ((double)nnz -nnzap)*((double)nnz -nnzap);
  nnz   = i;
  MPI_Allreduce(&nnzt,&nnzs,1,MPI_DOUBLE,MPI_SUM,A0->comm);
  nnzs  = (nnzs / (double)nprocs)/nnzap;
  MPI_Allreduce(&np,&i,1,MPI_INT,MPI_SUM,A0->comm);
  np = i;
#endif

  err = lis_vector_duplicate(A0,&x);
  if( err ) CHKERR(err);
  err = lis_vector_duplicate(A0,&b);
  if( err ) CHKERR(err);

  lis_matrix_get_range(A0,&is,&ie);
  for(i=0;i<n;i++)
    {
      val = ((double)(i+is+1))/(double)gn;
      val = 1.0;
      err = lis_vector_set_value(LIS_INS_VALUE,i+is,val,x);
    }
  for(i=0;i<n;i++)
    {
      lis_sort_id(A0->ptr[i],A0->ptr[i+1]-1,A0->index,A0->value);
    }
		
  rmin = gn;
  rmax = 0;
  rs   = 0;
  ra   = (double)nnz/(double)n;
  for(i=0;i<n;i++)
    {
      rn = A0->ptr[i+1] - A0->ptr[i];
      rb = A0->index[A0->ptr[i+1]-1] - A0->index[A0->ptr[i]];
      if( rb<rmin ) rmin = rb;
      if( rb>rmax ) rmax = rb;
      rs += ((double)rn - ra)*((double)rn - ra);
    }
  rs = rs/(double)n;
		
  if( my_rank==0 )
    {
      printf("n=%d nnz=%d nnzs=%8.1f np=%d mmin=%d mmax=%d\n",gn,nnz,nnzs,np,rmin,rmax);
    }

  time = lis_wtime();
  lis_matrix_duplicate(A0,&A);
  lis_matrix_set_type(A,storage);
  err = lis_matrix_convert(A0,A);
  time = lis_wtime() - time;
  convtime = time;
  if( my_rank==0 )
    {
      printf("convert time = %e\n",time);
    }
  if( err ) CHKERR(err);
		    
  comptime[0] = 100;
  comptime[1] = 100;
  for(i=0;i<iter;i++)
    {
#ifdef USE_MPI
      MPI_Barrier(A->comm);
#endif
      time = lis_wtime();
#ifdef USE_MPI
      time2 = lis_wtime();
      lis_send_recv(A->commtable,x->value);
      time2 = lis_wtime() - time2;
#endif
      lis_matvec(A,x,b);
#ifdef USE_MPI
      MPI_Barrier(A->comm);
#endif
      time = lis_wtime() - time;
      if( time<comptime[0] )
	{
	  comptime[0] = time;
	  commtime[0] = time2;
	}
    }
  lis_vector_nrm2(b,&val);

  for(i=0;i<iter;i++)
    {
#ifdef USE_MPI
      MPI_Barrier(A->comm);
#endif
      time = lis_wtime();
      lis_matvec(A,x,b);
#ifdef USE_MPI
      time2 = lis_wtime();
      lis_reduce(A->commtable,b->value);
      time2 = lis_wtime() - time2;
#endif
      time = lis_wtime() - time;
      if( time<comptime[1] )
	{
	  comptime[1] = time;
	  commtime[1] = time2;
	}
    }
  lis_vector_nrm2(b,&val2);

  if( my_rank==0 )
    {
      flops[0] = 2.0*nnz*1.0e-6 / comptime[0];
      flops[1] = 2.0*nnz*1.0e-6 / comptime[1];
	  printf("format = %s (%2d)  comptime1 = %e  MFLOPS = %8.3f commtime1 = %e nrm2 = %e\n",lis_storagename2[storage-1],storage,comptime[0],flops[0],commtime[0],val);
	  printf("format = %s (%2d)  comptime2 = %e  MFLOPS = %8.3f commtime2 = %e nrm2 = %e\n",lis_storagename2[storage-1],storage,comptime[1],flops[1],commtime[1],val2);
    }
  lis_matrix_destroy(A);
  lis_matrix_destroy(A0);
  lis_vector_destroy(b);
  lis_vector_destroy(x);

  lis_finalize();

  LIS_DEBUG_FUNC_OUT;

  return 0;
}


